------------------------------------------------------------------------------------
-- MIT License                                                                    --
------------------------------------------------------------------------------------
--                                                                                --
-- Copyright (c) 2024 apicici                                                     --
--                                                                                --
-- Permission is hereby granted, free of charge, to any person obtaining a copy   --
-- of this software and associated documentation files (the "Software"), to deal  --
-- in the Software without restriction, including without limitation the rights   --
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell      --
-- copies of the Software, and to permit persons to whom the Software is          --
-- furnished to do so, subject to the following conditions:                       --
--                                                                                --
-- The above copyright notice and this permission notice shall be included in all --
-- copies or substantial portions of the Software.                                --
--                                                                                --
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR     --
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,       --
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE    --
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER         --
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,  --
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE  --
-- SOFTWARE.                                                                      --
------------------------------------------------------------------------------------

--[[-- A pure Lua library to manage dialogues in point & click adventure games and other
dialogue-driven genres.
Provides facilities for managing dialogue trees, and through the included @{strands}
module can also be used as a system to manage cutscenes.

See @{plato_demo.lua} for a demonstration of how to use it.
@module plato
@alias M
]]

local path = (...):gsub("[^%.]*$", "")
local strands = require(path .. "strands")

local M = {}
--- Library name.
M._NAME="plato"
--- Library version.
M._VERSION="0.1.0"

local strand
local choices
local selected
local dialogue
local state
local env

local callbacks = {}
M.callbacks = callbacks

local states = {}
local initialised = {}

--[[-- Outer functions.
Functions meant to be run outside of a dialogue.
@section outer
]]

--[[-- Starts a new dialogue.
Cannot be used from within another dialogue (for that use @{divert}).
@see divert
@param t table describing the dialogue (see @{10-table_format.md|here} for details on the format of the table)
@tparam string entry_point name of the entry of `t`. Defaults to `"main"` if not provided
@raise
* no dialogue must be running
* `t` must be a table and `entry_point` must be a strong or nil
* `t` must have the `entry_point` field (or "main" if not specified)

@usage
    local t = {
        main = function()
            plato.option(1, "option 1", "exit")
            plato.option(2, "option 2", "exit")
            plato.selection()
        end,

        exit = function()
            -- do nothing, it's just an example!
        end
    }

    -- start the dialogue in the "main" entry point
    plato.start(t)
]]

function M.start(t, entry_point)
    assert(strand == nil, "Another dialogue is running.")
    if entry_point == nil then entry_point = "main" end
    assert(type(t) == "table", "First argument must be a table.")
    assert(type(entry_point) == "string", "Second argument must be a string.")

    dialogue = t

    local callback = callbacks.start
    if callback then callback(t, entry_point) end

    choices = {}

    if not states[t] then
        states[t] = {once={}, option_once={}, other={}}
    end
    state = states[t]

    if not initialised[t] then
        -- only do this if the first time we load this dialogue
        initialised[t] = true
        if env then
            for k, f in pairs(t) do
                if type(f) == "function" then
                    strands._setfenv(f, env)
                end
            end
        end
    end

    local f = assert(t[entry_point], string.format("Entry point '%s' missing.", entry_point))
    strand = strands.Strand.new(function()
        f()
        M.exit()
    end, env)
    strand:resume()
end

--[[-- Updates plato's internal @{strands|strand}. Does nothing if there is no dialogue running.
@number dt time elapsed since last update.
Only important if calling strands's @{strands.wait|wait} function from within the dialogue, in which case
the units should be chosen consistently.
@raise If there are no choices or the index is out of bounds
]]

function M.update(dt)
    if strand then strand:update(dt) end
end

--[[-- Get a reference to the table with the current dialogue choices, if any.
The array entries of the table are the text passed to
@{option} or @{option_once}. Additionally, the table contains a `"custom"` field
which is an array of the extra arguments passed to @{selection}.

This is the table that should be used, when the dialogue is running, to display the dialogue choices.

@treturn[1] table array containing the text of the dialogue choices, if a dialogue is running and there are choices.
@treturn[2] nil otherwise
]]

function M.get_choices()
    if choices and #choices > 0 then
        if not choices.out then
            local t = {}
            for i, c in ipairs(choices) do
                t[i] = c.text
            end
            t.custom = choices.custom
            choices.out = t
        end
        return choices.out
    end
end

--[[-- Select a dialogue choice.
@int i index of the selected choice
@raise if there are no choices or the index is out of bounds
]]

function M.select(i)
    local c = assert(choices and choices[i], "Invalid choice.")
    choices = {}
    selected = c
    strand:resume()
end

--[[-- Stops the currently running dialogue *from the outside*.
Useful for resetting the dialogue system when loading a savegame.
Resets everything but the dialogue states and the strand environment.
]]

function M.clear()
    strand = nil
    choices = nil
    selected = nil
    dialogue = nil
end

--[[-- Get the state table associated to a dialogue.
The format of the table is described @{20-state_format.md|here}.
@tparam table t table describing the dialogue
@treturn[1] table state table if the dialogue has been started or a table for
this dialogue has already been set through @{set_state_table}
@treturn[1] nil otherwise

]]

function M.get_state_table(t)
    return states[t]
end

--[[-- Set the state table associated to a dialogue.
The table must in the right format, described @{20-state_format.md|here}.
@tparam table t table describing the dialogue
@tparam ?table v state table. If not provided the state will be cleared
@raise If provided, `v` must have the correct format
@see check_state_table
]]

function M.set_state_table(t, v)
    if v == nil then
        states[t] = nil
        return
    end
    assert(M.check_state_table(v), "Provided table does not have the correct format.")
    states[t] = v
end

--[[-- Check if a table has the correct format for a state table.
Useful to ensure a table is suitable for @{set_state_table}, i.e. it
conforms to the appropriate  @{20-state_format.md|format}.
@tparam table v table describing the dialogue
@treturn bool `true` if the format is correct, `false` otherwise
]]

function M.check_state_table(v)
    return type(v) == "table" and
           type(v.once) == "table" and
           type(v.option_once) == "table" and
           type(v.other)
end

--[[-- Specify the environment to use for the fields of the dialogue tables.
See @{strand_environments.md|Strand environments} for more details.

By default `plato` does not change the function environments.

**Note**: the change of environment only affects dialogues that are started for the first time after
this function is called.

@tparam table|nil t table to use as the environment. Pass `nil` to clear the setting and revert to the default
@see get_environment
]]

function M.set_environment(t)
    env = t
end

--[[-- Get the current environment table.
Returns nil unless the environment table has been set through @{set_environment}.

@treturn[1] table currently set environment table if any
@treturn[2] nil otherwise
@see set_environment
]]

function M.get_environment()
    return env
end

--[[-- Check if there is a dialogue currently running.
@treturn bool `true` if running, `false` otherwise
]]

function M.is_running()
    return not not strand
end


--[[-- Inner functions.
Functions meant to be called inside a dialogue.
@section inner
]]

--- Stop the current dialogue.

function M.exit()
    M.clear()
    local callback = callbacks.exit
    if callback then callback() end
    coroutine.yield()
end

local function option(i, text, f, once_name, ...)
    if not choices[i] then
        local t = {text=text, f=f, once_name=once_name}
        choices[i] = t

        local extra = {}
        for k = 1, select("#", ...) do
            extra[k] = select(k, ...)
        end
        t.extra = extra
    end
end

--[[-- Add a dialogue option.
Whether the option will actually be used depends on its *weight*:
only the first viable option with the given weight will be presented
as a choice in @{get_choices}. See @{15-weight.md|here} for more information on weights.

**Note**: options don't need to be called in the order of their weights—only the order within
each weight matters, the winning options are then sorted before being presented according to
their weight. Likewise, weights don't have to be contiguous or start from any given number.

@see option_once
@int i weight of the option
@param text associated to the option.
Can be of any type. Not processed by plato,
will be passed as is when using @{get_choices}
@tparam ?string|function f
* if a string this is the entry point of the dialogue table to jump to if the option is selected
* if a function this will be called (from withing the dialogue strand) if the option is selected.
Useful for short consequences that don't require their own entry point
@param ... additional parameters for custom behaviour. Put sequentially in an array that can be accessed
using the @{callbacks.select} callback
]]

function M.option(i, text, f, ...)
    option(i, text, f, nil, ...)
end

--[[-- Add a dialogue option selectable only once.
The option will be skipped and the next weighted one will be chosen if it has
already been selected. The state is saved in the current dialogue @{20-state_format.md|state table}.
@see option
@param i same as @{option}
@string id a unique identifier (withing the current dialogue) for saving
the state of whether the dialogue has been selected. `option_once` state is
saved in its own subtable, so the name can be resused with @{once} and @{set_state}
without conflicts.
@param text same as @{option}
@param f same as @{option}
@param ... same as @{option}
]]

function M.option_once(i, id, text, f, ...)
    if state.option_once[id] then return end
    option(i, text, f, id, ...)
end

--[[-- Puts together the options and returns control
to the game while waiting for a choice.
*Must* be used after all the appropriate @{option} and @{option_once}
have been called for the dialogue system to work properly.

@param ... extra parameters that will be passed on the the table return by @{get_choices}

@usage

    t = {
        main = function()
            plato.option_once(1, "stayed", "stay", "main")
            plato.option(2, "leave", "exit")
            plato.selection()
        end,

        exit = function()
            plato.exit()
        end
    }
]]

function M.selection(...)
    if not choices then return end
    local tmp = choices
    choices = {}
    local indices = {}
    for k in pairs(tmp) do
        indices[#indices + 1] = k
    end

    if #indices == 0 then M.exit() end

    table.sort(indices)
    for i, k in ipairs(indices) do
        choices[i] = tmp[k]
    end

    local custom = {}
    for i = 1, select("#", ...) do
        custom[i] = select(i, ...)
    end
    choices.custom = custom

    while #choices > 0 do
        coroutine.yield()
    end

    if not selected then M.exit() end

    if selected.once_name then
        state.option_once[selected.once_name] = true
    end

    local callback = callbacks.select
    if callback then callback(selected.text, selected.extra) end

    local f = selected.f
    if type(f) == "function" then
        f()
    elseif type(f) == "string" then
        f = assert(dialogue[f], string.format("Entry point '%s' missing.", f))
        f()
    end

    M.exit()
end

--[[-- Jump to a different entry point of the dialogue table.
Used to chain different sections of the dialogue, as well as allowing
looping by jumping back to a previous point.

@string name name of the entry point to jump to
@raise the entry point must exist

@usage

    -- don't do this, the dialogue will be stuck in an infinite loop!
    local t = {
        main = function()
            plato.go("next")
        end(),

        ["next"] = function()
            plato.go("main")
        end()
    }
]]

function M.go(name)
    local f = assert(dialogue[name], string.format("Entry point '%s' missing.", name))
    f()
    M.exit()
end

--[[-- Return true only when called for the first time with a given name in the current dialogue.
Useful to make some things happen only once, e.g. only showing an option once, only giving a particular
answer the first time an option is chose.

The state is saved in the dialogue's @{20-state_format.md|state table}.

**Note**: the state changes *only if* the function is run, so be careful
if combining with other conditions using `and` and `or` as there may be short-circuiting
if not used as the first condition.

@string name the identifier used when saving the state of whether this has been run already.
@usage
    -- only show the next option once (skipped if we go
    -- through the dialogue again of we go back one step
    -- and repeat)
    if plato.once("test") then
        option(1, "option text", "next")
    end
]]

function M.once(name)
    local t = state.once
    if t[name] then return false end
    t[name] = true
    return true
end

--[[-- Stop the current dialogue and jump to a new one.
@see start
@param t same as @{start}
@param[opt] entry_point same as @{start}
]]

function M.divert(t, entry_point)
    M.clear()
    local callback = callbacks.exit
    if callback then callback() end
    M.start(t, entry_point)
    coroutine.yield()
end

--[[-- Get a value from the state table associated to the dialogue.
Useful to retrieve values saved with @{set_state}.
@see set_state

@param key key to get the value of. Can be of any non-nil type
]]

function M.get_state(key)
    return state.other[key]
end

--[[-- Save a value in the state table associated to the dialogue.
Useful to save some variables related to the dialogue (e.g., whether a question has been asked)
that has no use outside of it.
@see get_state_table
@see get_state

@param key key to associate the value to. Can be of any non-nil type
@param value value to set. Can be any type, including `nil` to unset the key
]]

function M.set_state(key, value)
    state.other[key] = value
end

--[[-- Callbacks.
Functions that can be set by the user to customise the behaviour of the library.

These can be set by setting a function of the appropriate format to the `plato.callbacks` table.
@section callbacks

@usage
    callbacks.exit = function()
        -- do stuff
    end
]]

--[[-- Function to call when exiting a dialogue.
Called by @{exit} and @{divert} before stopping the dialogue.
@function callbacks.exit
]]

--[[-- Function to call when starting a dialogue.
Called by @{start} before starting the dialogue.
@function callbacks.start
@param t the dialogue table passed to @{start}
@param entry_point the entry point passed to @{start}, or `"main"` if none was given
]]

--[[-- Function to call after selecting a dialogue choice.
Called by @{selection} after a choice has been made and before running the function/jumping to the entry point
specified by the option.

Useful for example if the player should speak the selected choice before continuing.
@function callbacks.select

@param text text that was passed to the option that was selected
@tparam table extra an array containing the extra values passed to the option that was selected
@see option
@see option_once

]]

return M